package com.superversion.rasa.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.util.StrUtil;
import com.superversion.rasa.config.ThreadPoolExecutorConfig;
import com.superversion.rasa.pojo.vo.RasaArgument;
import com.superversion.rasa.service.RasaCmdService;
import com.supervision.model.RasaModelInfo;
import com.supervision.model.User;
import com.supervision.service.RasaModeService;
import com.supervision.service.UserService;
import com.supervision.service.impl.UserServiceImpl;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.omg.CORBA.TRANSACTION_MODE;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.*;
import java.util.function.Predicate;

@Service
@Slf4j
@RequiredArgsConstructor
public class RasaCmdServiceImpl implements RasaCmdService {


    @Value("${rasa.models-path}")
    private String modelsPath;

    @Value("${rasa.endpoints}")
    private String endpoints;

    @Value("${rasa.config}")
    private String config;

    @Value("${rasa.data-path}")
    private String dataPath;


    @Value("${rasa.shell-env:/bin/bash}")
    private String shellEnv;

    @Value("${rasa.train-shell}")
    private String trainShell;

    @Value("${rasa.run-shell}")
    private String runShell;

    @Value("${rasa.shell.work:/home/rasa_manage/}")
    private String shellWork;

    private final UserService userService;

    @Override
    public String trainExec(RasaArgument argument) throws ExecutionException, InterruptedException, TimeoutException {

        String domain = dataPath+"domain.yml";
        List<String> cmds = ListUtil.toList(shellEnv, trainShell,config,dataPath,domain,modelsPath);
        cmds.add(argument.getFixedModelName());

        log.info("trainExec cmd : {}",StrUtil.join(" ",cmds));

        List<String> outMessage = execCmd(cmds, s -> false, 300);


        boolean isSuccess = trainIsSuccess(outMessage);
        return String.join("\r\n",outMessage);
    }

    private  boolean trainIsSuccess(List<String> messageList){

        if (CollectionUtil.isEmpty(messageList)){
            return false;
        }
        String keyWord =  "Your Rasa model is trained and saved at";
        if (StrUtil.isEmpty(keyWord)){
            return false;
        }
       return messageList.stream().anyMatch(s->StrUtil.isNotEmpty(s) && s.contains(keyWord));
    }




    @Override
    public String runExec(RasaArgument argument) throws ExecutionException, InterruptedException, TimeoutException {


        String mPath = modelsPath+argument.getFixedModelName()+".tar.gz";
        List<String> cmds = ListUtil.toList(shellEnv, runShell,mPath,endpoints,argument.getPort());

        log.info("runExec cmd : {}",StrUtil.join(" ",cmds));
        return String.join("\r\n",execCmd(cmds,s-> StrUtil.isNotBlank(s)&& s.contains("Rasa server is up and running"),90));
    }

    @Override
    public void test() {
        List<User> list = userService.list();
        //RasaModelInfo rasaModelInfo = rasaModeService.queryByModelId("1");

        System.out.println(list);
    }


    private List<String> execCmd(List<String> cmds, Predicate<String> endPredicate, long timeOut) throws InterruptedException, ExecutionException, TimeoutException {
        ProcessBuilder processBuilder = new ProcessBuilder(cmds);
        processBuilder.directory(new File(shellWork));
        processBuilder.redirectErrorStream(true);

        ThreadPoolExecutor instance = ThreadPoolExecutorConfig.getInstance();
        Future<List<String>> future = instance.submit(() -> {
            Process process = processBuilder.start();
            InputStream inputStream = process.getInputStream();
            Reader reader = new InputStreamReader(inputStream, "UTF-8");
            BufferedReader bufferedReader = new BufferedReader(reader);
            List<String> outString = new ArrayList<>();
            String resultLines = bufferedReader.readLine();
            while( resultLines != null) {
                resultLines = bufferedReader.readLine();
                log.info("resultLines:{}",resultLines);
                outString.add(resultLines);
                if (endPredicate.test(resultLines)){
                    break;
                }
            }
            bufferedReader.close();
            return outString;
        });

        return future.get(timeOut, TimeUnit.SECONDS);
    }

}
